/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *     
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/


import java.util.*;
import java.io.*;

/** This class holds an abstract description of methods, and is able to 
    emit both Java wrapper, C headers and C glue code.
    
    Bugs: JNI Name mangling is not implemented. Please avoid
    underscore in method names.  Will eventually be written because
    of the naming changes requested by miguel.
    
    @author Olivier Gutknecht */

public class MethodDefinition {
    boolean debug = false;
    String nativeName;
    String methodName;
    String nativePreGlue = null;
    String nativeGlue = null;
    String nativePostGlue = null;
    TypeDefinition returnType;
    TypeDefinition objectType;
    String objectName; // FIXME should be accessed with the class definition
    String packageName; // FIXME should be accessed with the class definition
    
    boolean isStatic = false;
    Vector args;
    
    public static String convertFunc(String s, String p, String o) {
	//System.out.println(s+"-"+p+"-"+o);
	
	StringTokenizer st = new StringTokenizer(s, "_");
	String obj = new String(o);
	String mid;
	if ((Parser.naming == Parser.PURE_JAVA) || o.equals("")) {
	    mid = st.nextToken();
	    //System.out.println(" "+mid+" "+p.toLowerCase());
	    if (!mid.equals(p.toLowerCase()))
		return null;
	}
	while (st.hasMoreTokens() && (obj.length() > 0)) {
	    mid = st.nextToken();
	    // System.err.println("an "+mid+"/"+obj);
	    if (obj.toLowerCase().startsWith(mid))
		obj = obj.substring(mid.length());
	    else
		return null;
	}
	if (!st.hasMoreTokens())
	    return null;
	StringBuffer methodName = new StringBuffer(st.nextToken());
	if (Parser.naming != Parser.PURE_C) {
	    while (st.hasMoreTokens()) {
		mid = st.nextToken();
		methodName.append((mid.substring(0, 1)).toUpperCase() + 
				  mid.substring(1, mid.length()));
	    }
	}
	else {
	    while (st.hasMoreTokens()) {
		methodName.append("_" + st.nextToken());
	    }
	}
	return new String(methodName);
    }

    public void emitJavaDefaultStubs(PrintWriter pw, int arguments, int defaultValues) {
	// This section create the methods for the methods who has default values
	if (defaultValues > 0)
	    pw.println("\n  /* " + methodName + " has " + defaultValues + 
		       " default argument value(s) */");
	for (int i = 1; i <= defaultValues; i++) {
	    // First writes the prototype(stub) method
	    int defaultValueNo = 0;
	    pw.print("  ");
	    if (isStatic)
		pw.print("static ");
	    pw.print("public ");
	    pw.print(returnType.getJavaType() + " " + methodName + "(");
	    Enumeration e = args.elements();
	    boolean firstIsDefault = false;
	    Argument a = (Argument) e.nextElement();
	    if (!(a.hasDefaultValue() && i == defaultValues))
		pw.print(a.toString());
	    else
		firstIsDefault = true;
	    if (a.hasDefaultValue())
		defaultValueNo++;
	    for (int j = 1; j < arguments; j++) {
		a = (Argument) e.nextElement();
		if (a.hasDefaultValue())
		    defaultValueNo++;
		if (!(a.hasDefaultValue() && defaultValueNo > defaultValues - i)) {
		    if (firstIsDefault)
			firstIsDefault = false;
		    else
			pw.print(", ");
		    pw.print(a.toString());
		}
	    }
	    pw.println(") {");
	    
	    // Then write the call to the real method with the default arguments
	    defaultValueNo = 0;
	    pw.print("    ");
	    pw.print(methodName + "(");
	    e = args.elements();
	    a = (Argument) e.nextElement();
	    if (!(a.hasDefaultValue() && i == defaultValues))
		pw.print(a.getName());
	    else
		pw.print(a.getDefaultValue());
	    if (a.hasDefaultValue())
		defaultValueNo++;
	    for (int j = 1; j < arguments; j++) {
		a = (Argument) e.nextElement();
		if (a.hasDefaultValue())
		    defaultValueNo++;
		pw.print(", ");
		if (!(a.hasDefaultValue() && defaultValueNo > defaultValues - i))
		    pw.print(a.getName());
		else
		    pw.print(a.getDefaultValue());
	    }
	    pw.println(");\n  }\n");
	}
	pw.println("");
    }

    public void emitJavaMethod(PrintWriter pw) {
	String javaret = returnType.getJavaType();

	// emit a definition for a TYPE static variable if this is the
	// method "static GtkType getType()"
	if (isStatic &&
	    javaret.equals("GtkType") &&
	    methodName.equals("getType") &&
	    (args.size() == 0)) {
	    pw.println("  static public GtkType TYPE = getType();");
	}

	// whether or not to do a wrapped native method
	boolean wrapNative = mustWrapNative();
	
	// figure out the artument list, in a few variants, note argument
	// count, and count default args
	//
	// javaArgs:   "(int x, GtkWidget y,           GtkCellType z)"
	// unwrapArgs: "(    x,           y,                       z.getValue()"
	// nativeArgs: "(int x, GtkWidget y, int /* GtkCellType */ z);
	StringBuffer javaArgs = new StringBuffer();
	StringBuffer unwrapArgs = new StringBuffer();
	StringBuffer nativeArgs = new StringBuffer();
	javaArgs.append('(');
	unwrapArgs.append('(');
	nativeArgs.append('(');
	int defaultValues = 0;
	int arguments = 0;
	Enumeration e = args.elements();
	while (e.hasMoreElements()) {
	    Argument a = (Argument) e.nextElement();
	    if (a.hasDefaultValue())
		defaultValues++;
	    if (arguments != 0) {
		javaArgs.append(", ");
		unwrapArgs.append(", ");
		nativeArgs.append(", ");
	    }
	    if (a.getType().isEnumValue()) {
		// FIXME: there should be a method on type definition to return
		// the right thing for this case
		nativeArgs.append("int /* ");
		nativeArgs.append(a.getType().getJavaType());
		nativeArgs.append(" */ ");
		nativeArgs.append(a.getName());
		unwrapArgs.append(a.getName());
		unwrapArgs.append(".getValue()");
	    }
	    else {
		nativeArgs.append(a.toString());
		unwrapArgs.append(a.getName());
	    }
	    javaArgs.append(a.toString());
	    arguments++;
	}
	javaArgs.append(')');
	unwrapArgs.append(')');
	nativeArgs.append(')');

	if (wrapNative) {
	    // we determined that we need to output a wrapper for the native
	    // function; write the native method as "<methodName>0" and the
	    // wrapper as the original method name
	    pw.print("  native public ");
	    if (isStatic)
		pw.print("static ");
	    if (returnType.isEnumValue()) {
		// FIXME: there should be a method on type definition to return
		// the right thing for this case
		pw.print("int /* " + returnType.getJavaType() + " */");
	    }
	    else
		pw.print(javaret);
	    pw.print(" ");
	    pw.print(methodName);
	    pw.print("0");
	    pw.print(nativeArgs.toString());
	    pw.println(";");

	    pw.print("  public ");
	    if (isStatic)
		pw.print("static ");
	    pw.print(javaret);
	    pw.print(" ");
	    pw.print(methodName);
	    pw.print(javaArgs.toString());
	    pw.println(" {");

	    String call0 = methodName + "0" + unwrapArgs.toString();
	    
	    if (returnType.isEnumValue()) {
		pw.print("    return ");
		pw.print(javaret);
		pw.print(".intern(");
		pw.print(call0);
		pw.println(");");
	    }
	    else if (returnType.getJavaType().equals("void")) {
		pw.print("    ");
		pw.print(call0);
		pw.println(";");
	    }
	    else {
		pw.print("    return ");
		pw.print(call0);
		pw.print(";");
	    }
	    pw.println("  }");
	}
	else {
	    // unwrapped native method
	    pw.print("  native public ");
	    if (isStatic)
		pw.print("static ");
	    pw.print(javaret);
	    pw.print(" ");
	    pw.print(methodName);
	    pw.print(javaArgs.toString());
	    pw.println(";");
	}

	emitJavaDefaultStubs(pw, arguments, defaultValues);
    }

    public void emitNativeBody(PrintWriter pw) {
	int i;
	if (nativePreGlue != null) {
	    pw.print(nativePreGlue + '\n');
	}
	
	/* Local variable declarations */
	
	if (!isStatic) {
	    /* To get the object itself native pointer */
	    pw.println("  " + objectType.getNativeType() + " cptr;");
	    pw.println("  jfieldID fid;");
	    pw.println("  jclass cls;");
	    pw.println("  jlong jptr;");
	}
	
	/* Local variables for argument translation */
	i = 0;
	for (Enumeration e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitNativeDeclaration(pw, "_arg" + i);
	    i++;
	}
	
	/* If the native api returns something, we should also prepare the translation */
	if (!returnType.isVoid() && !returnType.isTime_t()) {
	    pw.print("  " + returnType.getNativeType());
	    pw.println(" _result;");
	    pw.print("  " + returnType.getJNIType());
	    pw.println(" _jresult;");
	}
	else if (returnType.isTime_t()) {
	    pw.print("  " + returnType.getNativeType());
	    pw.println(" _result;");
	    pw.println("  jmethodID mid_jresult;");
	    pw.println("  jclass cls_jresult;");
	    pw.print("  " + returnType.getJNIType());
	    pw.println(" _jresult;");
	}
	pw.println();
	
	/* JNI to native casts */
	if (!isStatic) {
	    /* Getting native object pointer */
	    pw.println("  cls = (*env)->GetObjectClass(env, obj);");
	    pw.println("  fid = (*env)->GetFieldID(env, cls, \"nativepeer\", \"J\");");
	    pw.println("  jptr = (*env)->GetLongField(env, obj, fid);");
	    pw.println("  cptr = *(" + objectType.getNativeType() + "*)&jptr;");
	    pw.println();
	}
	
	/* Argument translation */
	i = 0;
	for (Enumeration e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitJavaToNativeConversion(pw, "_arg" + i);
	    i++;
	}
	if (debug)
	    pw.println("  printf(\"Native code in: " + methodName + "\\n\");");
	Enumeration e = args.elements();
	i = 0;
	pw.print("  ");
	if (nativeGlue != null) {
	    pw.print(nativeGlue + '\n');
	}
	else {
	    if (!returnType.isVoid())
		pw.print("_result = (" + returnType.getNativeType() + ")");
	    if (isStatic) {
		pw.print(nativeName + "(");
		if (e.hasMoreElements()) {
		    pw.print("_arg" + i);
		    e.nextElement();
		    i++;
		}
	    }
	    else
		pw.print(nativeName + "(cptr");
	    for (; e.hasMoreElements();) {
		e.nextElement();
		pw.print(", _arg" + i);
		i++;
	    }
	    pw.println(");");
	}
	i = 0;
	for (e = args.elements(); e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    a.emitNativeToJavaConversion(pw, "_arg" + i);
	    i++;
	}
	if (nativePostGlue != null) {
	    pw.print(nativePostGlue + '\n');
	}
	if (!returnType.isVoid() && !returnType.isTime_t()) {
	    pw.println("  _jresult = (" + returnType.getJNIType() + ")_result;");
	    pw.println("  return _jresult;");
	}
	if (returnType.isTime_t()) {
	    pw.println("  cls_jresult = (*env)->FindClass(env, \"java/util/Date\");");
	    pw.println("  mid_jresult = (*env)->GetMethodID(env, cls_jresult, \"<init>\", \"(J)V\");");
	    pw.println("  _jresult = (*env)->NewObject(env, cls_jresult, mid_jresult, ((jlong)_result)*1000);");
	    pw.println("  return _jresult;");
	}
    }

    public void emitNativeComment(PrintWriter pw) {
	pw.println("/*");
	pw.println(" * Class:     " + packageName + "_" + objectName);
	pw.println(" * Method:    " + methodName);
	pw.println(" * Signature: " + getSignature());
	pw.println(" */");
    }

    public void emitNativeHeader(PrintWriter pw) {
	emitNativeComment(pw);
	pw.print("JNIEXPORT " + returnType.getJNIType());
	pw.print(" JNICALL " + "Java_" + packageName + "_" + objectName + 
		 "_" + methodName);
	
	// add '0' to the name if it's going to be wrapped
	if (mustWrapNative())
	    pw.print("0");

	pw.print(" (JNIEnv *, ");
	if (isStatic)
	    pw.print("jclass");
	else
	    pw.print("jobject");
	Enumeration e = args.elements();
	for (; e.hasMoreElements();) {
	    pw.print(", ");
	    pw.print(((Argument) e.nextElement()).getType().getJNIType());
	}
	pw.println(");");
	pw.println();
    }
    /**
     * Get the value of nativeGlue.
     * @return Value of nativeGlue.
     */

    public String getNativeGlue() {
	return nativeGlue;
    }

    public String getSignature() {
	StringBuffer sb = new StringBuffer("(");
	Enumeration e = args.elements();
	for (; e.hasMoreElements();) {
	    sb.append(((Argument) e.nextElement()).getType().getSignature());
	}
	sb.append(")" + returnType.getSignature());
	return sb.toString();
    }

    public MethodDefinition(String n, TypeDefinition r, Vector a) {
	nativeName = n;
	returnType = r;
	args = a;
	packageName = "";
	methodName = "";
    }

    public void printNativeFunction(PrintWriter pw) {
	emitNativeComment(pw);
	pw.print("JNIEXPORT ");
	pw.print(returnType.getJNIType() + " JNICALL ");
	pw.print("Java_" + packageName + "_" + objectName + "_" + methodName);

	// add '0' to the name if it's goint to be wrapped
	if (mustWrapNative())
	    pw.print("0");
	
	pw.println();

	pw.print("  (JNIEnv * env, ");
	if (isStatic)
	    pw.print("jclass cls");
	else
	    pw.print("jobject obj");
	Enumeration e = args.elements();
	for (; e.hasMoreElements();) {
	    Argument a = (Argument) e.nextElement();
	    pw.print(", ");
	    pw.print(a.getType().getJNIType());
	    pw.print(" " + "j_" + a.getName());
	}
	pw.println(")");
	pw.println("{");
	emitNativeBody(pw);
	pw.println("}");
	pw.println();
    }

    public void setNativeGlue(String v) {
	this.nativeGlue = v;
    }

    public void setNativePostGlue(String v) {
	this.nativePostGlue = v;
    }

    /**
     * Set the value of nativeGlue.
     * @param v  Value to assign to nativeGlue.
     */
    public void setNativePreGlue(String v) {
	this.nativePreGlue = v;
    }

    public void setObject(TypeDefinition o) {
	objectType = o;
    }

    public boolean translate(String prefix, TypeDefinition obj) {
	String m;
	if (obj != null) {
	    m = convertFunc(nativeName, Parser.currentModule, obj.getJavaType());
	    //System.err.println(nativeName+" / "+obj.getRawValue()+" "+obj.getJavaType()+m);
	    objectType = obj;
	}
	else
	    m = convertFunc(nativeName, Parser.currentModule, "");
	if (m != null) {
	    methodName = m;
	    if (obj == null)
		objectName = prefix;
	    else {
		objectName = obj.getJavaType();
	    }
	    packageName = prefix.toLowerCase();
	    
	    /* removing the first argument if similar to object parameter */
	    if (args.size() > 0) {
		Argument first = (Argument) args.firstElement();
		if (first.getType().equals(obj)) {
		    //System.err.println(first.getType()+" is equal to "+obj);
		    args.removeElementAt(0);
		}
		else {
		    //   System.err.println(first.getType()+" not equal to "+obj);
		    isStatic = true;
		}
	    }
	    else
		isStatic = true;
	    return true;
	}
	else
	    return false;
    }

    /**
     * Return <code>true</code> if the native method needs to be wrapped.
     * Wrapping happens when the return type or any arguments are of an
     * enumerated type
     *
     * @return true if the native method needs to be wrapped
     */
    private boolean mustWrapNative() {
	if (returnType.isEnumValue())
	    return true;
	Enumeration e = args.elements();
	while (e.hasMoreElements()) {
	    Argument a = (Argument) e.nextElement();
	    if (a.getType().isEnumValue())
		return true;
	}
	return false;
    }
}
